/*-
 * Copyright (C)2000, 2001, 2002 @BABOLO http://www.links.ru/~babolo/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2000, 2001, 2002 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: getCGIparm.c,v 1.12 2003/11/04 08:37:39 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <stdio.h>
#include <time.h>
#include <err.h>
#include "getCGIparm.h"

extern char *optarg;
extern int optind;
extern int optopt;
extern int opterr;
extern int optreset;

void
usage(int ex) {
    int i;
    printf("getCGIparm " VERS "  " DATE " @BABOLO http://www.links.ru/~babolo/\n"
           "getCGIparm [Hnrptvwxz][-c text][-f file][-q quoten] type [name]\n"
           "known patterns:"
          );
    for (i = 0; getCGIparmpatterns[i].name; i++)
        printf(" %s", getCGIparmpatterns[i].name);
    printf("\nknown quotens:");
    for (i = 0; getCGIparmquotens[i].name; i++)
        printf(" %s", getCGIparmquotens[i].name);
    printf("\n"
           " -a retry find nonempty parameter\n"
           " -c error text\n"
           " -f error message file\n"
           " -e error messages file\n"
           " -H use SERVER_PROTOCOL in addition to QUERY_STRING\n"
           " -n delete LF\n"
           " -r no CR\n"
           " -p pipe from stdin\n"
           " -t test mode - no output\n"
           " -v verbose\n"
           " -w alternate &\n"
           " -x[x[x]] strictness levels\n"
           " -z text instead empty parameter\n"
          );
    exit(ex);
};

int
main(int argc, char **argv) {
#define INBUF 1048576
    u_char *query = getenv("QUERY_STRING");
    u_char *proto = getenv("SERVER_PROTOCOL");
    u_char *unescaped, *p, *q;
    int i, j, ex = EX_OK, sillyloop;
    char ch;
    const u_char *empty = "";
    u_char *errform, *quot, *zero;
    u_int32_t flags;
    u_char *query_delimit;

#ifdef DEBUG
# if DEBUG > 1
    getCGIparmdebug1();
# endif
#endif
#ifdef SPY
    {   FILE *spy;
        const u_char *empty = "";
        int i;
# ifdef DEBUG
        spy = stderr;
# else
        spy = fopen(SPY, "a");
# endif
        if  (spy) {
            for (i = 0; i < argc; i++)
                fprintf(spy, "%s~", argv[i]);
            fprintf( spy, "\nQUERY_STRING=%s~\nSERVER_PROTOCOL=%s~\n"
                   , query ? query : empty
                   , proto ? proto : empty
                   );
# ifndef DEBUG
            fclose(spy);
# endif
    }   }
#endif
    flags = 0;
    quot = NULL;
    zero = NULL;
    errform = NULL;
    query_delimit = NULL;
    while ((ch = getopt(argc, argv, "ac:e:f:hHnpq:rtvw:xz:")) != EOF)
        switch (ch) {
         case 'a': flags |= gCGI_RTRY;            break;
         case 'c': errform = optarg;
                   flags &= ~gCGI_OUTMASK;
                   flags |= gCGI_TEXT;
                                                  break;
         case 'e': errform = optarg;
                   flags &= ~gCGI_OUTMASK;
                   flags |= gCGI_FORM;
                                                  break;
         case 'f': errform = optarg;
                   flags &= ~gCGI_OUTMASK;
                   flags |= gCGI_FILE;
                                                  break;
         case 'h': usage(ex);                     break;
         case 'H': if  (proto == NULL) {
                       if  (flags & gCGI_VERB) warnx("SERVER_PROTOCOL absent");
                       exit(EX_USAGE);
                   }
                   if  (strncmp(proto, "HTTP", 4)) {
                       q = malloc(strlen(query) + strlen(proto) + 2);
                       strcpy(q, query);
                       strcat(q, " ");
                       strcat(q, proto);
                       p = strstr(q, " HTTP");
                       if (p) *p = '\0';
                       query = q;
                   };                             break;
         case 'p': query = malloc(INBUF);
                   q = query; i = INBUF - 1;
                   /* It is possible that apache infinitely
                    * do writes with 0 length to pipe.
                      Restrict by sillyloop and usleep().
                    */
                   sillyloop = 3;
                   while (feof(stdin) == 0 && sillyloop > 0 && i > 0) {
                       j = fread(q, 1, i, stdin);
                       q += j; i -= j;
                       if  (j) {
                           sillyloop = 3;
                       } else {
                           sillyloop--;
#ifdef SPY
                           if  (sillyloop == 0) {
                               FILE *spy;
                               time_t t;
# ifdef DEBUG
                               spy = stderr;
# else
                               spy = fopen(SPY, "a");
# endif
                               if  (spy) {
                                   t = time(NULL);
                                   fprintf(spy, "sillyloop %s", ctime(&t));
# ifndef DEBUG
                                   fclose(spy);
# endif
                           }   }
#endif
                           if  (sillyloop) usleep(500);
                   }   };
                   while (*--q < ' ') {};
                   *++q = '\0';
                                                  break;
         case 'q': quot = optarg;                 break;
         case 'n': flags |= gCGI_NOLF;            break;
         case 'r': flags |= gCGI_NOCR;            break;
         case 't': flags |= gCGI_OTST;            break;
         case 'v': flags |= (flags << 1) & gCGI_VERM;
                   flags |= gCGI_VERB;
                                                  break;
         case 'w': query_delimit = optarg;        break;
         case 'x': flags |= (flags << 1) & gCGI_ERRMASK;
                   flags |= gCGI_DOER;
                                                  break;
         case 'z': flags |= gCGI_ZERO;
                   zero = optarg;
                                                  break;
         default:  usage(EX_USAGE);
        }
    argc -= optind;
    argv += optind;
    if  (flags & gCGI_VERB) {
        for (i = 0; i < argc; i++) fprintf(stderr, "%s~", argv[i]);
        fprintf( stderr, "\nQUERY_STRING=%s~\nSERVER_PROTOCOL=%s~\n"
               , query ? query : empty
               , proto ? proto : empty
               );
    }

    if  (argc == 0) usage(EX_USAGE);
    if  (argc > 0 && !argv[0]) usage(EX_USAGE);
    if  (argc > 1 && !argv[1]) usage(EX_USAGE);
    if  (argc > 2) usage(EX_USAGE);
    if  (!query) query = "";
    unescaped = malloc(strlen(query) * (query_delimit ? strlen(query_delimit) : 1) + 1);
    if  (!unescaped)                                                       errx(EX_TEMPFAIL, "malloc");
    p = unescaped;
    i = 0;

    for (;;) {
        ex = getCGIparmpass1(flags, &query, argv[1], p);
        if  (flags & gCGI_VERB) fprintf(stderr, "pass1 = %d\n", ex);
        if  (ex || *query++ != '&') break;
        if  (query_delimit) p = index(strcat(unescaped, query_delimit), '\0');
        else if (flags & gCGI_RTRY && !*unescaped) i++;
        else break;
    }

    if  (ex && i) {
        *unescaped = '\0';
        ex = EX_OK;
    }

    if  (!ex) {
        ex = getCGIparmpass2(flags, argv[0], unescaped);
        if  (!ex) {
            ex = getCGIparmpass3(flags, quot, &unescaped);
            if  (!ex && flags & gCGI_NONL && !*unescaped) ex = EX_DATAERR;
        } else if (~flags & gCGI_MATC) {
            *unescaped = '\0';
            ex = EX_OK;
        }
    } else if (~flags & gCGI_DOER) ex = EX_OK;

    if  (ex) getCGIparmpass4(flags, argv[0], argv[1], errform, stderr);
    else if (~flags & gCGI_OTST) printf("%s", (flags & gCGI_ZERO && !*unescaped) ? zero : unescaped);

#ifdef SPY
    {   FILE *spy;
        time_t t;
# ifdef DEBUG
        spy = stderr;
# else
        spy = fopen(SPY, "a");
# endif
        if  (spy) {
            t = time(NULL);
            fprintf( spy, "%04X, %2d, OUT=%s~\n%s\n"
                   , flags, ex, (flags & gCGI_ZERO && !*unescaped) ? zero : unescaped, ctime(&t)
                   );
# ifndef DEBUG
            fclose(spy);
# endif
    }   }
#endif
    exit(ex);
}
