/*-
 * Copyright (C)2000, 2001, 2002 @BABOLO http://www.links.ru/~babolo/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2000, 2001, 2002 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: pass3.c,v 1.6 2003/11/04 08:37:39 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "getCGIparm_int.h"
#include "getCGIparm.h"

static const int
quote_literal(const u_char *pattern, const u_char *query, u_char **result) {
    char *p;
    *result = malloc(strlen(query) * 2 + 1);
    p = *result;
    for (; *query; query++)
    {   switch(*query)
        {case '\'': case '\\': *p++ = *query; *p++ = *query; break;
         default: *p++ = *query;
    }   }
    *p++ = '\0';
    return(EX_OK);
};

static const int
quote_ident(const u_char *pattern, const u_char *query, u_char **result) {
    char *p;
    *result = malloc(strlen(query) * 2 + 1);
    p = *result;
    for (; *query; query++)
    {   switch(*query)
        {case '\"': case '\\': *p++ = *query; *p++ = *query; break;
         default: *p++ = *query;
    }   }
    *p++ = '\0';
    return(EX_OK);
};

static const int
quote_single(const u_char *pattern, const u_char *query, u_char **result) {
    char *p;
    *result = malloc(strlen(query) + 3);
    p = *result;
    strcpy(p, "\'");
    strcat(p, query);
    strcat(p, "\'");
    return(EX_OK);
};

static const int
quote_double(const u_char *pattern, const u_char *query, u_char **result) {
    char *p;
    *result = malloc(strlen(query) + 3);
    p = *result;
    strcpy(p, "\"");
    strcat(p, query);
    strcat(p, "\"");
    return(EX_OK);
};

const struct getCGIparmquotens getCGIparmquotens[] =
{ {"psql" , 0, quote_literal}
, {"psqi" , 0, quote_ident  }
, {"sq"   , 0, quote_single }
, {"dq"   , 0, quote_double }
, {NULL   , 0, NULL         }
};

int
getCGIparmpass3( u_int32_t flags
     , const u_char *q
     , u_char **unescaped
     ) {
    int i, ex;
    u_char *r;
    const u_char * const empty = "";

#ifdef DEBUG
    fprintf(stderr, "%04X pass3 ~%s~%s~\n", flags, q, *unescaped);
#endif
    if  (flags & gCGI_VERB)
        fprintf( stderr
               , "P2 flags=%06X, q=%s~, unescaped=%s~\n"
               , flags
               , q ? q : empty
               , unescaped ? (*unescaped ? *unescaped : empty) : empty
               );
    ex = EX_OK;
    if  (!unescaped)
        ex = EX_SOFTWARE;
    else if (!q || !*q) ex = EX_OK;
    else if (~flags & gCGI_ZERO || **unescaped)
    {   u_char *p = NULL;

        while (!ex && q && *q)
        {   ex = EX_USAGE;
            r = strsep(&q, "/");
            for (i = 0; getCGIparmquotens[i].name; i++)
            {   if  ( getCGIparmquotens[i].namelen
                    ? !strncmp(getCGIparmquotens[i].name, r, getCGIparmquotens[i].namelen)
                    : !strcmp(getCGIparmquotens[i].name, r)
                    )
                {   if  (!(ex = (getCGIparmquotens[i].test_proc)(r, *unescaped, &p) ? EX_DATAERR : EX_OK))
                    {   if  (*unescaped != p)
                        {   free(*unescaped);
                            *unescaped = p;
                            if  (flags & gCGI_VERB) fprintf(stderr, "%d NEW=%s~\n", ex, p);
                        }else
                        {   if  (flags & gCGI_VERB) fprintf(stderr, "%d OLD=%s~\n", ex, p);
                    }   }
                    break;
    }   }   }   }
    return(ex);
}
